//-----------------------------------------------------------------------------
// QUEUE.CPP
// By Dean Harding
// 
// Copyright (c) 2000/2001 Dean Harding
//-----------------------------------------------------------------------------
#include "internal.h"

//-----------------------------------------------------------------------------
template< class T >
inline CQueue<T>::CQueue()
: m_pTop( NULL ), m_pBottom( NULL ), m_hQueueLock( NULL ), m_hAddEvent( NULL )
, m_hEmptyEvent( NULL )
{
	m_hQueueLock = CreateMutex( NULL, FALSE, NULL );
	m_hAddEvent = CreateEvent( NULL, FALSE, FALSE, NULL );
	m_hEmptyEvent = CreateEvent( NULL, FALSE, FALSE, NULL );
}
//-----------------------------------------------------------------------------
template< class T >
inline CQueue<T>::~CQueue()
{
	// if the queue is not empty, that's bad!
	ASSERT( m_pTop == NULL && m_pBottom == NULL );

	CloseHandle( m_hEmptyEvent );
	CloseHandle( m_hAddEvent );
	CloseHandle( m_hQueueLock );
}

//-----------------------------------------------------------------------------
template< class T >
inline T* CQueue<T>::GetNext( bool bBlock )
{
	ASSERT( WaitForSingleObject( m_hQueueLock, INFINITE ) == WAIT_OBJECT_0 );

	// if it's empty and we're not blocking, then just return NULL
	if( IsEmpty() && bBlock == false )
	{
		return NULL;
	}

	// otherwise, if it's empty, we have to block until there's something there
	// we know that's happened when m_hItemAdded is set.
	else if( IsEmpty() )
	{
		ReleaseMutex( m_hQueueLock );
		ASSERT( WaitForSingleObject( m_hAddEvent, INFINITE ) == WAIT_OBJECT_0 );
		ASSERT( WaitForSingleObject( m_hQueueLock, INFINITE ) == WAIT_OBJECT_0 );
	}

	// therwise, it's not empty, so make sure we reset any events that are
	// signaled
	else
	{
		ResetEvent( m_hAddEvent );
	}

	SQueueData *pData = NULL;

	// now we can be sure there's at least one item in the list.
	// if there's *only* one, then we have to lock the queue so that another
	// thread doesn't try and add something to the end while we remove the
	// only item that's there.
	if( m_pTop == m_pBottom )
	{
//		ASSERT( WaitForSingleObject( m_hQueueLock, INFINITE ) == WAIT_OBJECT_0 );

		pData = m_pTop;
		m_pTop = m_pBottom = NULL;

//		ReleaseMutex( m_hQueueLock );

		// now there's *no* items in the list, so we signal the empty event
		SetEvent( m_hEmptyEvent );
	}
	else
	{
		// we can remove stuff from the top without worrying about what any
		// other threads are doing to the bottom!
		pData = m_pTop;
		m_pTop = m_pTop->pNext;
	}

	ReleaseMutex( m_hQueueLock );

	// now pData should point to the top of the list!
	ASSERT( pData != NULL );
	T* pItem = pData->pData;

	delete pData;
	return pItem;
}

//-----------------------------------------------------------------------------
template< class T >
inline void CQueue<T>::AddItem( T* pItem )
{
	// we do as much as possible before we actually lock the bottom of the queue
	// that way, the queue is locked for as short a time as possible.
	SQueueData *pData = new SQueueData;
	pData->pData = pItem;
	pData->pNext = NULL;

	// we have to wait for the m_hBottomLock mutex so that we know that no one
	// else is fiddling with the bottom of the list
	ASSERT( WaitForSingleObject( m_hQueueLock, INFINITE ) == WAIT_OBJECT_0 );

	// now we insert pData onto the end of the list
	if( m_pBottom ) m_pBottom->pNext = pData;
	pData->pPrev = m_pBottom;
	m_pBottom = pData;

	if( m_pTop == NULL ) m_pTop = m_pBottom;

	// make sure we reset the empty event, because the queue is no longer empty!
	ResetEvent( m_hEmptyEvent );
	
	// tell any threads waiting on the BottomLock mutex that it's now OK to
	// proceed.
	ReleaseMutex( m_hQueueLock );

	// now that we've added an item, set the m_hItemAdded event so that any
	// thread blocking on GetNext will wake up
	SetEvent( m_hAddEvent );
}

//-----------------------------------------------------------------------------
template< class T >
inline bool CQueue<T>::IsEmpty()
{
	return( m_pTop == NULL );
}

//-----------------------------------------------------------------------------
template< class T >
inline void CQueue<T>::WaitForEmpty()
{
	while( !IsEmpty() )
		Sleep( 10 );
//	ASSERT( WaitForSingleObject( m_hEmptyEvent, INFINITE ) == WAIT_OBJECT_0 );

//	Sleep(10000);
}

//-----------------------------------------------------------------------------
// End of QUEUE.CPP
//-----------------------------------------------------------------------------
