//-----------------------------------------------------------------------------
// WORKER.CPP
// By Dean Harding
// 
// Copyright (c) 2000/2001 Dean Harding
//-----------------------------------------------------------------------------
// This is the thread that actually does all the work.  It waits on the
// queue for more items, whenever a new item is added, it writes that item
// to a file!
//-----------------------------------------------------------------------------
#include "internal.h"

//-----------------------------------------------------------------------------
CQueue<CLoggerItem>g_LoggerQueue;

//-----------------------------------------------------------------------------
// While this lock is set, no threads from other processes will try and open
// the file.  This is for things like when renaming a large file or such
HANDLE g_hFileLock = NULL;
const TCHAR g_szFileLock[] = _T("dhLoggerFileLock");

//-----------------------------------------------------------------------------
// renames an open file, and returns a handle to a new file of the same name.
HANDLE RenameFile( HANDLE hFile, TCHAR *szFileName );

//-----------------------------------------------------------------------------
// opens a file.  If access is denied, keep trying
HANDLE OpenFile( TCHAR *szFileName );

//-----------------------------------------------------------------------------
DWORD WINAPI LoggerWorkerThread( void *pData )
{
	TCHAR szBuffer[1024];
	DWORD dwNumWritten;

	g_hFileLock = CreateMutex( NULL, FALSE, g_szFileLock );
	
	while( true )
	{
		CLoggerItem *pItem = g_LoggerQueue.GetNext( true );

		// if it's a "NULL" item, then that's our queue to quit!
		if( pItem == NULL )
			break;

		HANDLE hFile = OpenFile( pItem->szLogFile );

		DWORD dwFileSize = GetFileSize( hFile, NULL );
		if( dwFileSize > (256 * 1024) )
		{
			// if the file is bigger than 256KB, then rename it and make a new
			// one
			hFile = RenameFile( hFile, pItem->szLogFile );
		}

		// seek to the end of the file
		SetFilePointer( hFile, 0, 0, FILE_END );

		// if they want a timestamp, format that for them
		if( (pItem->dwFormat & LOG_TIMESTAMP) != 0 )
		{
			FILETIME ftLocalTime;
			SYSTEMTIME stLocalTime;

			FileTimeToLocalFileTime( &pItem->ftTimestamp, &ftLocalTime );
			FileTimeToSystemTime( &ftLocalTime, &stLocalTime );

			_stprintf( szBuffer, _T("%2d/%02d/%04d %2d:%02d:%02d.%03d "),
				stLocalTime.wDay, stLocalTime.wMonth, stLocalTime.wYear,
				stLocalTime.wHour, stLocalTime.wMinute, stLocalTime.wSecond,
				stLocalTime.wMilliseconds );

			WriteFile( hFile, szBuffer, _tcslen( szBuffer ), &dwNumWritten, NULL );
		}

		// if they want a process name & thread id
		if( (pItem->dwFormat & LOG_PROCNAME) != 0 && (pItem->dwFormat & LOG_THREADID) != 0 )
		{
			_stprintf( szBuffer, _T("(%s:%d) "), pItem->szProcName, pItem->dwThreadId );

			WriteFile( hFile, szBuffer, _tcslen( szBuffer ), &dwNumWritten, NULL );
		}
		// if it's just the process name
		else if( (pItem->dwFormat & LOG_PROCNAME) != 0 )
		{
			_stprintf( szBuffer, _T("(%s)"), pItem->szProcName );

			WriteFile( hFile, szBuffer, _tcslen( szBuffer ), &dwNumWritten, NULL );
		}
		// if it's just the thread id
		else if( (pItem->dwFormat & LOG_THREADID) != 0 )
		{
			_stprintf( szBuffer, _T("(%d)"), pItem->dwThreadId );

			WriteFile( hFile, szBuffer, _tcslen( szBuffer ), &dwNumWritten, NULL );
		}

		// now write the actual line
		WriteFile( hFile, pItem->szLine, _tcslen( pItem->szLine ), &dwNumWritten, NULL );

		// and an end-of-line
		WriteFile( hFile, _T("\r\n"), _tcslen( _T("\r\n") ), &dwNumWritten, NULL );

		// and close the file!
		CloseHandle( hFile );

		delete pItem;
	}

	CloseHandle( g_hFileLock );

	return 0;
}

//-----------------------------------------------------------------------------
HANDLE RenameFile( HANDLE hFile, TCHAR *szFileName )
{
	ASSERT( WaitForSingleObject( g_hFileLock, INFINITE ) == WAIT_OBJECT_0 );

	// now that we own the file lock, we can close the file, rename and make a
	// new one!!

	CloseHandle( hFile );

	TCHAR szBuffer[MAX_PATH];
	TCHAR szNewFile[MAX_PATH];
	FILETIME ftCurrent, ftLocal;
	SYSTEMTIME stLocal;

	TCHAR *p = (TCHAR *)(szFileName + _tcslen(szFileName) );
	while( *(p - 1) != '\\' && p != szFileName ) p--;

	GetSystemTimeAsFileTime( &ftCurrent );
	FileTimeToLocalFileTime( &ftCurrent, &ftLocal );
	FileTimeToSystemTime( &ftLocal, &stLocal );

	_stprintf( szBuffer, _T("%04d-%02d-%2d %02d-%02d-%02d %s"),
		stLocal.wYear, stLocal.wMonth, stLocal.wDay,
		stLocal.wHour, stLocal.wMinute, stLocal.wSecond,
		p );

	_tcsncpy( szNewFile, szFileName, (p - szFileName) );
	p = szNewFile + (p - szFileName);
	*p = '\0';

	_tcscat( szNewFile, szBuffer );

	MoveFile( szFileName, szNewFile );

	ReleaseMutex( g_hFileLock );
	return OpenFile( szFileName );
}

//-----------------------------------------------------------------------------
HANDLE OpenFile( TCHAR *szFileName )
{
	// open the file, keep trying until it's not INVALID_HANDLE_VALUE
	// There has to be a better way than this.  I mean, I want to be able
	// to block the thread until the file is available again.  I don't know
	// if there's any API for that...
	HANDLE hFile = INVALID_HANDLE_VALUE;
	while( hFile == INVALID_HANDLE_VALUE )
	{
		// we only try to open the file if we own the hFileLock mutex.
		ASSERT( WaitForSingleObject( g_hFileLock, INFINITE ) == WAIT_OBJECT_0 );

		hFile = CreateFile(
			szFileName,
			GENERIC_WRITE,
			FILE_SHARE_READ,
			NULL,
			OPEN_ALWAYS,
			FILE_ATTRIBUTE_NORMAL,
			NULL );
		Sleep( 0 );

		ReleaseMutex( g_hFileLock );
	}

	return hFile;
}

//-----------------------------------------------------------------------------
// End of WORKER.CPP
//-----------------------------------------------------------------------------
